/***********************************************************************************************

    VocProc
    Copyright (C) 2010 Igor Brkić <igor@hyperglitch.com>

    phase vocoder code taken from:
    http://www.dspdimension.com/admin/pitch-shifting-using-the-ft/


    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

***********************************************************************************************/


#include "vproc_dsp.h"
#include <QtDebug>


void mydsp::phaseVocAnalysis(fftw_complex *block, float *gLastPhase, double freqPerBin, double expct, float *gAnaMagn, float *gAnaFreq){

    double real, imag, magn, phase, tmp;
    long qpd, k;

    for (k = 0; k <= fftFrameSize/2; k++) {

        /* de-interlace FFT buffer */
        real = block[k][0];
        imag = block[k][1];

        /* compute magnitude and phase */
        magn = 2.*sqrt(real*real + imag*imag);
        phase = atan2(imag,real);

        /* compute phase difference */
        tmp = phase - gLastPhase[k];
        gLastPhase[k] = phase;

        /* subtract expected phase difference */
        tmp -= (double)k*expct;

        /* map delta phase into +/- Pi interval */
        qpd = tmp/M_PI;
        if (qpd >= 0) qpd += qpd&1;
        else qpd -= qpd&1;
        tmp -= M_PI*(double)qpd;

        /* get deviation from bin frequency from the +/- Pi interval */
        tmp = overlap*tmp/(M_2PI);

        /* compute the k-th partials' true frequency */
        tmp = (double)k*freqPerBin + tmp*freqPerBin;

        /* store magnitude and true frequency in analysis arrays */
        gAnaMagn[k] = magn;
        gAnaFreq[k] = tmp;
    }
}

void mydsp::phaseVocSynthesis(fftw_complex *block, float *gSumPhase, float *gSynMagn, float *gSynFreq, double freqPerBin, double expct){

    int k;
    double magn, tmp, phase;

    for (k = 0; k <= fftFrameSize/2; k++) {
        /* get magnitude and true frequency from synthesis arrays */
        magn = gSynMagn[k];
        tmp = gSynFreq[k];

        /* subtract bin mid frequency */
        tmp -= (double)k*freqPerBin;

        /* get bin deviation from freq deviation */
        tmp /= freqPerBin;

        /* take overlap into account */
        tmp = M_2PI*tmp/overlap;

        /* add the overlap phase advance back in */
        tmp += (double)k*expct;

        /* accumulate delta phase to get bin phase */
        gSumPhase[k] += tmp;
        phase = gSumPhase[k];

        /* get real and imag part and re-interleave */
        block[k][0] = magn*cos(phase);
        block[k][1] = magn*sin(phase);
    }
}

void mydsp::spectralEnvelope(float *env, fftw_complex *fft, int count){

    int nTaps=20;
    int nTaps2=10;
    float tmp[count+nTaps];

    float h[]={ // h=(firls(20, [0 0.02 0.1 1], [1 1 0 0]));
        0.0180, 0.0243, 0.0310, 0.0378, 0.0445, 0.0508, 0.0564, 0.0611,
        0.0646, 0.0667, 0.0675, 0.0667, 0.0646, 0.0611, 0.0564, 0.0508,
        0.0445, 0.0378, 0.0310, 0.0243, 0.0180
    };
    
    // |H(w)|
    memset(tmp+count, 0, nTaps);
    for(int k=0;k<count;k++)
        tmp[k]=sqrt(fft[k][0]*fft[k][0]+fft[k][1]*fft[k][1]);

    memset(env, 0, count*sizeof(float));

    // magnitude spectrum filtering
    int i, j;
    float *p_h, *p_z, accum;

    float z[2 * nTaps];
    memset(z, 0, 2*nTaps*sizeof(float));
    int state = 0;
    for (j = 0; j < count+nTaps2; j++) {
        z[state] = z[state + nTaps] = tmp[j];
        p_h = h;
        p_z = z + state;
        accum = 0;
        for (i = 0; i < nTaps; i++) accum += *p_h++ * *p_z++;
        if (--state < 0) state += nTaps;
        if(j>=nTaps2) env[j-nTaps2]=accum;
    }
}

float mydsp::pitchFrequency(fftw_complex *block){
    // cepstral method - needs improvement but good for now

    float freq;
    int i;
    double cepst[fftFrameSize/2];
    int ppMin, ppMax;
    double max1;
    float pitch=0;

    //double fftTmpReal[fftFrameSize];
    //fftw_complex fftTmpCplx[fftFrameSize];

    for(i=0;i<fftFrameSize/2;i++){
        fftCeps[i][0]=log(sqrt(pow(block[i][0], 2)+pow(block[i][1], 2))+1e-6)/(float)fftFrameSize;
        fftCeps[i][1]=0;
    }

    fftPlan=fftw_plan_dft_c2r_1d(fftFrameSize, fftCeps, fftTmpR, FFTW_ESTIMATE);
    fftw_execute(fftPlan);
    fftw_destroy_plan(fftPlan);

    // normalize
    for(i=0;i<fftFrameSize/2;i++) cepst[i]=fabs(fftTmpR[i]/(float)fftFrameSize)+1e6;

    ppMax=fftFrameSize/2-2;
    ppMin=fSamplingFreq/1200; // fMax=1200

    // find maximum of HTP cepstrum
    max1=0;
    for(i=ppMin;i<=ppMax;i++){
        if(cepst[i]>max1){
            max1=cepst[i];
            pitch=i;
        }
    }

    // interpolate between two samples
    int idx=pitch;
    int l1;

    if(cepst[idx-1]>cepst[idx+1]) l1=pitch-1;
    else l1=pitch;

    pitch=l1+1/(cepst[l1]/cepst[l1+1]+1);

    freq=fSamplingFreq/pitch;

    return freq;
}

void mydsp::setPitchFactor(float freq){
    // find nearest tone from chosen scale

    /*
      scale[i] - i-th scale in XML file
      scale[i][0] - number of elements in i-th scale
    */

    float *curr=NULL;
    int i;
    int sc=(int)sScale;
    for(i=2;i<scales[sc][1];i++){
        curr=&scales[sc][i];
        if(freq<scales[sc][i]) break;
    }

    // add transposition if needed (and if possible)
    if((i+sTranspose)<=scales[sc][0] && (i+sTranspose)>=1){
        curr+=(int)sTranspose;
    }

    // add hysteresis to reduce oscillation around middle point

    float sign;
    if((freqOld-freq)>0) sign=-1; // pitch is falling
    else sign=1; // pitch is rising

    float thr=(*(curr-1)+*curr)/2 + 0.3*sign*(*curr-*(curr-1));

    // if current frequency is lower than threshold use lower note
    if(freq<thr) curr--;

    // see if correction is needed at all
    if((fabs(*curr-freq)/freq)>sThreshold){
        // do pitch factor averaging (attack)
        sPitchFactor*=(1+(float)((int)(sAttack*20.0)));
        sPitchFactor+=(*curr/freq);
        sPitchFactor/=((float)((int)(sAttack*20.0))+2.0);
    }
    else
        sPitchFactor=1;

    // honor limits
    if(sPitchFactor>2.0 || sPitchFactor<0.5) sPitchFactor=1.0;

}

void mydsp::compute (int count, float** input, float** output) {
    float* input0 = input[0];
    float* input1 = input[1];
    float* output0 = output[0];

    static float gLastPhase[MAX_FRAME_LENGTH/2+1];
    static float gSumPhase[MAX_FRAME_LENGTH/2+1];
    static float gAnaFreq[MAX_FRAME_LENGTH];
    static float gAnaMagn[MAX_FRAME_LENGTH];
    static float gSynFreq[MAX_FRAME_LENGTH];
    static float gSynMagn[MAX_FRAME_LENGTH];

    static long gRover = false, gInit = false;

    double freqPerBin, expct;
    long i,k, index, inFifoLatency, stepSize, fftFrameSize2;

    float *fPointer, *fPointer2, *fPointer3;
    double *dPointer;

    /* set up some handy variables */
    fftFrameSize2 = fftFrameSize/2;

    stepSize = fftFrameSize/overlap;
    freqPerBin = (double)fSamplingFreq/(double)fftFrameSize;
    expct = M_2PI*(double)stepSize/(double)fftFrameSize;

    inFifoLatency = fftFrameSize-stepSize;
    if (gRover == false) gRover = inFifoLatency;

    /* initialize our static arrays */
    if (gInit == false) {
        memset(gLastPhase, 0, (MAX_FRAME_LENGTH/2+1)*sizeof(float));
        memset(gSumPhase, 0, (MAX_FRAME_LENGTH/2+1)*sizeof(float));
        memset(gAnaFreq, 0, MAX_FRAME_LENGTH*sizeof(float));
        memset(gAnaMagn, 0, MAX_FRAME_LENGTH*sizeof(float));
        gInit = true;
    }

    /* main processing loop */
    for (i = 0; i < count; i++){

        // As long as we have not yet collected enough data just read in
        gInFIFO[gRover] = input0[i];
        gIn2FIFO[gRover] = input1[i];
        output0[i] = gOutFIFO[gRover-inFifoLatency];
        gRover++;

        // now we have enough data for processing
        if (gRover >= fftFrameSize) {
            gRover = inFifoLatency;

            float tmpPower=0.0;
            dPointer=fftTmpR;
            fPointer=gInFIFO;
            fPointer2=window;
            for (k = 0; k < fftFrameSize;k++) {
                *dPointer=*(fPointer++) * *(fPointer2++);
                tmpPower+= *dPointer * *dPointer;
                dPointer++;
            }

            powerIn=tmpPower*100;

            // do transform
            fftPlan=fftw_plan_dft_r2c_1d(fftFrameSize, fftTmpR, fftTmpC, FFTW_ESTIMATE);
            fftw_execute(fftPlan);
            fftw_destroy_plan(fftPlan);

            // pitch correction
            if(cAutoTune){
                float freq;
                freq=pitchFrequency(fftTmpC);
                setPitchFactor(freq);
            }

            memcpy(fftOldC, fftTmpC, fftFrameSize*sizeof(fftw_complex));

            // pitch shifting with phase vocoder
            phaseVocAnalysis(fftTmpC, gLastPhase, freqPerBin, expct, gAnaMagn, gAnaFreq);
            memset(gSynMagn, 0, fftFrameSize*sizeof(float));
            memset(gSynFreq, 0, fftFrameSize*sizeof(float));
            for (k = 0; k <= fftFrameSize2; k++) {
                index = k*sPitchFactor;
                if (index <= fftFrameSize2) {
                    gSynMagn[index] += gAnaMagn[k];
                    gSynFreq[index] = gAnaFreq[k] * sPitchFactor;
                    if(cEffect)
                        gSynFreq[index] = gSynFreq[index]*sEffect + sEffect*200*(float)rand()/RAND_MAX-100;
                }
            }
            phaseVocSynthesis(fftTmpC, gSumPhase, gSynMagn, gSynFreq, freqPerBin, expct);

            // formant correction + vocoder
            if(cFormantVoco){
                float env1[fftFrameSize2], env2[fftFrameSize2];
                if(sSwitch){
                    dPointer=fftTmpR; fPointer=gIn2FIFO; fPointer2=window;
                    for (k = 0; k < fftFrameSize;k++) {
                        *dPointer++=*(fPointer++) * *(fPointer2++);
                    }

                    fftPlan=fftw_plan_dft_r2c_1d(fftFrameSize, fftTmpR, fftTmpC, FFTW_ESTIMATE);
                    fftw_execute(fftPlan);
                    fftw_destroy_plan(fftPlan);
                }

                spectralEnvelope(env1, fftOldC, fftFrameSize2);
                spectralEnvelope(env2, fftTmpC, fftFrameSize2);

                // modify spectral envelope of spectrum in fftTmpC to look like spectral
                // envelope of spectrum in fftOldC
                float koef;
                fPointer2=env1; fPointer3=env2;
                for(k=0;k<fftFrameSize2;k++){
                    koef = *(fPointer2++) / (*(fPointer3++)+.02)*2;
                    fftTmpC[k][0] *= koef;
                    fftTmpC[k][1] *= koef;
                }
            }

            // do inverse transform
            fftPlan=fftw_plan_dft_c2r_1d(fftFrameSize, fftTmpC, fftTmpR, FFTW_ESTIMATE);
            fftw_execute(fftPlan);
            fftw_destroy_plan(fftPlan);

            tmpPower=0.0;
            fPointer=gOutputAccum; dPointer=fftTmpR; fPointer2=window;
            for(k=0; k < fftFrameSize; k++) {
                *fPointer += 2.0 * *(dPointer++) / (fftFrameSize2*overlap) * sOutputGain * *(fPointer2++);
                tmpPower+= *fPointer * *fPointer;
                fPointer++;
            }

            powerOut=tmpPower*100;

            memcpy(gOutFIFO, gOutputAccum, stepSize*sizeof(float));

            // shift accumulator
            memmove(gOutputAccum, gOutputAccum+stepSize, fftFrameSize*sizeof(float));

            // move input FIFO
            for (k = 0; k < inFifoLatency; k++) gInFIFO[k] = gInFIFO[k+stepSize];
            for (k = 0; k < inFifoLatency; k++) gIn2FIFO[k] = gIn2FIFO[k+stepSize];
        }
    }
}

